import re
from pathlib import Path
import shutil
from typing import Generator, List, Callable


def _get_image_names(markdown_file_path: Path) -> List[str]:
    """
    Get image names from .md file

    Args:
        markdown_file_path (Path): md file path

    Returns:
        List[str]: image names in md file
    """
    image_names = []
    patt = (
        r"!\[\[(.*?)\.(png|jpg|jpeg|gif)\]\]"  # group 1 = file_name, group 2 = foramt
    )
    with markdown_file_path.open("r") as f:
        while line := f.readline():
            if search_results := re.findall(patt, line):
                for search_result in search_results:
                    image_name = f"{search_result[0]}.{search_result[1]}"  # extract abc.png like names
                    image_names.append(image_name)
    return image_names


def _get_all_images(image_folder_path: Path) -> dict[str, Path]:
    """
    get all images in folder and create [name, Path] dict of all iamges

    Args:
        image_folder_path (Path): image folder path

    Returns:
        dict[str, Path]: {name, Path}
    """
    patt = r".*\.(png|jpg|jpeg|gif)$"
    name_path_map: dict[str, Path] = {}
    # search all subfolders
    for file_path in image_folder_path.rglob("**/*"):
        if re.search(patt, file_path.suffix):
            name_path_map[file_path.name] = file_path
    return name_path_map


def create_image_path_generator(image_folder_path: Path) -> Callable:
    name_path_map: dict = _get_all_images(
        image_folder_path
    )  # image name, path dict of all images in folder

    def yield_image_path(markdown_file_path: Path) -> Generator[Path, None, None]:
        """
        Create Genrator yields image paths in md file

        Args:
            markdown_file_path:(Path): md file path

        Yields:
            Generator[Path, None, None]: yield image path generator
        """
        image_names: list[str] = _get_image_names(markdown_file_path)
        for image_name in image_names:
            if image_name in name_path_map:
                yield name_path_map[image_name]

    return yield_image_path


def _replace_name_to_url(line: str, image_name: str, s3_url: str):
    # Local file link -> S3 url
    local_image_link = r"!\[\[" + image_name + r"\]\]"
    if local_link := re.search(local_image_link, line):  # 이미지 링크가 존재하는지 탐색
        s3_link = f"![]({s3_url})"
        line = line.replace(local_link.group(), s3_link)  # s3 링크로 대체
    return line


def write_md_file(
    markdown_file_path: Path,
    output_folder_path: Path,
    link_replace_map: List[tuple[str, str]],
    is_overwrite: bool = False,
):
    """
    Write new .md that replace local file link to S3 url.
    Only replace image file link and other things are same
    Args:
        markdown_file_path (Path): md file path
        output_folder_path (Path): output file path
        link_replace_map (List[tuple[str, str]]): file link -> s3 url
    """
    out_file_path = output_folder_path.joinpath(markdown_file_path.name)
    with open(markdown_file_path, "r") as origin_file:  # open origin file
        with open(out_file_path, "w") as output_file:
            while line := origin_file.readline():
                for image_name, s3_url in link_replace_map:
                    line = _replace_name_to_url(line, image_name, s3_url)
                output_file.write(line)  # if no replace just copy line

    if is_overwrite:
        shutil.move(out_file_path, markdown_file_path)



import concurrent.futures
from pathlib import Path

from .markdown import create_image_path_generator, write_md_file
from .s3 import S3

from typing import Generator, Callable, List, Tuple


def put_images_in_md(
    s3: S3, markdown_path: Path, image_path_generator: Generator[Path, None, None]
) -> List[Path]:
    """
    Generate local images path by using generator and put images to S3.
    Return sccessfully put image paths

    Args:
        s3 (S3): instance to control S3
        markdown_file_name (str): makrdown file name
        image_path_generator (Generator[Path, None, None]): yield image paths to upload

    Returns:
        List[Path]: sccessfully put image paths
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
        futures = []
        for image_path in image_path_generator:
            futures.append(
                executor.submit(s3.put_image, markdown_path, image_path)
            )  # run upload by multithread

        put_image_paths: List[Path] = []  # put success image path list
        for future in concurrent.futures.as_completed(futures, timeout=60):
            try:
                put_image_paths.append(future.result())
            except Exception as e:
                print(e)
                continue

    return put_image_paths


def create_obs3dian_runner(
    s3: S3,
    image_folder_path: Path,
    output_folder_path: Path,
    is_overwrite: bool = False,
) -> Callable:
    """
    Create runner fucntion object
    S3 controller and ouput_folder_path would not change before config

    Args:
        s3 (S3): S3 controller
        image_folder_path (Path): image folder path
        output_folder_path (Path): output folder path

    Returns:
        Callable: runner
    """

    image_path_creator: Callable = create_image_path_generator(
        image_folder_path
    )  # function to make image path generator

    def run(markdown_file_path: Path) -> None:
        """
        Run obs3dian command extract image paths and replace them by S3 URLs.

        Args:
            markdown_file_path (Path): mark down file path
        """
        image_path_generator: Generator = image_path_creator(markdown_file_path)
        put_image_paths = put_images_in_md(s3, markdown_file_path, image_path_generator)

        # (image name, S3 URL) to convert link
        link_replace_pairs: List[Tuple[str, str]] = []
        for image_path in put_image_paths:
            image_name = image_path.name
            s3_url = s3.get_image_url(markdown_file_path, image_path)
            link_replace_pairs.append(
                (image_name, s3_url)
            )  # image name in .md would convert to S3 url

        write_md_file(
            markdown_file_path, output_folder_path, link_replace_pairs, is_overwrite
        )  # write new md with S3 link
        return

    return run



from dataclasses import dataclass
import json
from pathlib import Path
import typer

APP_NAME = "obs3dian"
SETUP_FILE_NAME = "config.json"
APP_DIR_PATH = typer.get_app_dir(APP_NAME)


@dataclass(frozen=True)
class Configuration:
    profile_name: str = ""
    aws_access_key: str = ""
    aws_secret_key: str = ""
    bucket_name: str = "obs3dian"
    output_folder_path: str = "./output"
    image_folder_path: str = "./images"


def load_configs() -> Configuration:
    """
    Load config json file from app dir

    Returns:
        dict: config json
    """

    try:
        config_path = Path(APP_DIR_PATH) / SETUP_FILE_NAME
        with config_path.open("r") as f:
            configs = json.load(f)
        return Configuration(**configs)  # create config data class

    except FileNotFoundError as e:
        print("Config file is not founded... input your configuration\n")
        raise e


def save_config(json_data: dict):
    app_dir_path = Path(APP_DIR_PATH)  # create app setting folder
    if not app_dir_path.exists():
        app_dir_path.mkdir()

    config_path = app_dir_path / SETUP_FILE_NAME
    with config_path.open("w") as f:
        json.dump(json_data, f)  # write config.json


def remove_config():
    app_dir_path = Path(APP_DIR_PATH)  # create app setting folder
    config_path = app_dir_path / SETUP_FILE_NAME
    try:
        config_path.unlink()
    except FileNotFoundError:
        print("Config file is not exists run config first")



import boto3
from botocore.exceptions import ClientError
import json
from pathlib import Path
from urllib import parse


class S3:
    """
    Class to control S3
    this class can make bucket, set public access, put images in s3.
    To use this class you need AWS CLI profile which has permission to create public bucket and put images.
    """

    def __init__(
        self,
        profile_name: str,
        bucket_name: str,
        aws_access_key: str,
        aws_secret_key: str,
        usekey: bool,
    ) -> None:

        if usekey:
            if aws_access_key and aws_secret_key:  # if user has cli profile
                self.session = boto3.Session(
                    aws_access_key_id=aws_access_key,
                    aws_secret_access_key=aws_secret_key,
                    region_name="ap-northeast-2",
                )
                
        elif profile_name:
            self.session = boto3.Session(profile_name=profile_name)
        
        else:
            raise ValueError("AWS key is required")

        self.s3 = self.session.client("s3")
        self.bucket_name = bucket_name
        return

    def _check_bucket_exist(self) -> bool:
        try:
            self.s3.head_bucket(Bucket=self.bucket_name)  # check bucket is exists
            return True

        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            if error_code == "404":  # if bucket not exists return false
                return False
            elif error_code == "403":  # 403 means bucket exists but no permission
                return True
            else:  # bad reqeust
                raise e

    def _put_public_access_policy(self) -> None:
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "id-1",
                    "Action": ["s3:GetObject"],
                    "Effect": "Allow",
                    "Resource": f"arn:aws:s3:::{self.bucket_name}/*",
                    "Principal": "*",
                }
            ],
        }
        try:
            self.s3.put_bucket_policy(
                Bucket=self.bucket_name, Policy=json.dumps(policy)
            )  # put bucket public access

        except Exception as e:
            print("Can't allow public acess to bucket")
            raise e

    def create_bucket(self) -> bool:
        try:
            if self._check_bucket_exist():  # If bucket already exists
                return False

            bucket_name = self.bucket_name
            self.s3.create_bucket(
                CreateBucketConfiguration={"LocationConstraint": "ap-northeast-2"},
                Bucket=bucket_name,
                ObjectOwnership="ObjectWriter",
            )

            self.s3.put_public_access_block(
                Bucket=bucket_name,
                PublicAccessBlockConfiguration={
                    "BlockPublicAcls": False,
                    "IgnorePublicAcls": False,
                    "BlockPublicPolicy": False,
                    "RestrictPublicBuckets": False,
                },
            )
            self._put_public_access_policy()
            return True

        except ClientError as e:
            print("Error occured in create bucket")
            raise e

    def get_image_url(self, markdown_path: Path, image_path: Path) -> str:
        region = self.session.region_name
        key = f"{markdown_path.stem} / {image_path.name}"
        s3_url = f"https://{self.bucket_name}.s3.{region}.amazonaws.com/{parse.quote(key)}"  # image uploaded url
        return s3_url

    def put_image(self, markdown_path: Path, image_path: Path) -> Path:
        try:
            self.s3.put_object(
                Bucket=self.bucket_name,
                Body=image_path.open("rb"),
                Key=f"{markdown_path.stem} / {image_path.name}",
                ContentType=f"image/{image_path.suffix}",
            )  # upload image
            return image_path

        except ClientError as e:
            print(f"Error Occured in uploading {image_path}")
            raise e



import concurrent.futures as concurrent_futures

import typer
from typing_extensions import Annotated

from pathlib import Path
import time

from .core import create_obs3dian_runner
from .config import load_configs, save_config, remove_config, APP_NAME, Configuration
from .s3 import S3


app = typer.Typer(name=APP_NAME)


def _convert_path_absoulte(path: Path, strict=True) -> Path:
    """

    Resolve path with ~ / ../ and convert it to absolute path

    Args:
        path (Path): path

    Returns:
        Path: Absolute path
    """
    try:
        resolved_path = path.expanduser().resolve(strict=strict)
        return resolved_path

    except OSError as e:
        print("Path is invalid ", e)
        raise e


def _render_animation(future: concurrent_futures.Future, echo_text: str) -> None:
    """
    Rendering loading animation

    Args:
        future (concurrent_futures.Future): future to wait
        description (str): description text to echo in terminal
    """
    i = 0
    animation = ["|", "/", "--", "\\"]
    while not future.done():  # run until uploading is done
        typer.echo(
            f"\r{animation[i % 4]:<2} {echo_text[:20]:<20}", nl=False
        )  # render animation like => | abc.md -> / abc.md -> -- abc.md
        time.sleep(0.2)  # give delay to show animation
        i += 1
    return


@app.command()
def apply(
    usekey: Annotated[
        bool,
        typer.Option(
            help="Use cli profile when connect to S3. If you want to connect by key uses --no-usekey"
        ),
    ] = True
):
    """
    Apply settings from config.json file

    Raises:
        e: invalid output path
    """
    configs: Configuration = load_configs()
    output_folder_path = _convert_path_absoulte(
        Path(configs.output_folder_path), strict=False
    )

    s3 = S3(
        profile_name=configs.profile_name,
        bucket_name=configs.bucket_name,
        aws_access_key=configs.aws_access_key,
        aws_secret_key=configs.aws_secret_key,
        usekey=usekey,  # use profile or key
    )

    if s3.create_bucket():
        print(f"Bucket {configs.bucket_name} created")
        print("Bucket has public read access so anyone can see files in your bucket")
    else:
        print(f"Bucket {configs.bucket_name} is already exists try with other name")
        return

    if not output_folder_path.exists():
        try:
            output_folder_path.mkdir()
            print(f"Create Output Folder in {output_folder_path.name}")

        except OSError as e:
            print(f"Output path {output_folder_path} is invalid")
            raise e
    else:
        print("Output Folder is already exists (Can modify by run obs3dian config)")


@app.command()
def config(
    reset: Annotated[
        bool,
        typer.Option(help="Reset config.json file"),
    ] = False,
):
    """
    Writes config file. This command sets setting for run obs3dian.\n
    It should be executed first before run obs3dian. If you want to apply setting use command 'apply'
    """
    if reset:
        remove_config()  # save empty data to reset config file
        return

    default_input = (
        lambda description, default: input(f"{description} [{default}]: ").strip()
        or default
    )
    try:
        # Last config data
        configs: Configuration = load_configs()

    except FileNotFoundError:
        configs: Configuration = Configuration()

    profile_name = configs.profile_name
    aws_access_key = configs.aws_access_key
    aws_secret_key = configs.aws_secret_key
    bucket_name = configs.bucket_name
    output_path = configs.output_folder_path
    image_folder_path = configs.image_folder_path

    typer.echo("AWS-CLI profile name or AWS key is required")
    if input("Do you want to config AWS-CLI profile name? (Y/N): ") in ["y", "Y"]:
        profile_name = default_input("AWS CLI Profile Name", profile_name)

    else:
        aws_access_key = default_input("AWS AccessKey", aws_access_key)
        aws_secret_key = default_input("AWS Secret Key", aws_secret_key)
        if not (aws_access_key and aws_secret_key):
            raise ValueError("You must provide both access and secret key")

    bucket_name = default_input("S3 bucket Name", bucket_name)

    output_path = Path(default_input("Output Path", output_path))
    output_path = _convert_path_absoulte(output_path, False)

    image_folder_path = Path(default_input("Image Folder Path", image_folder_path))
    image_folder_path = _convert_path_absoulte(image_folder_path, True)

    json_data = {
        "profile_name": profile_name,
        "aws_access_key": aws_access_key,
        "aws_secret_key": aws_secret_key,
        "bucket_name": bucket_name,
        "output_folder_path": str(output_path),
        "image_folder_path": str(image_folder_path),
    }

    save_config(json_data)


@app.command()
def run(
    user_input_path: Path,
    overwrite: Annotated[
        bool,
        typer.Option(
            help="Overwrites original md file in same file path. (default is creating new file under output folder)"
        ),
    ] = False,
    usekey: Annotated[
        bool,
        typer.Option(
            help="Use cli profile when connect to S3. If you want to connect by key uses --no-usekey"
        ),
    ] = False,
):
    """
    Get images local file paths from md files in given path.
    After extracting image paths, it uploads images to S3 and replaces all file links in .md to S3 links.
    Outputs will be wrriten under output folder and it's image links would replaced to S3 links

    Args:
        user_input_path (Path): your markdown file path to convert. (dir or file)
    """

    configs: Configuration = load_configs()

    apply(usekey)  # run apply
    typer.echo("")  # new line

    user_input_path = _convert_path_absoulte(user_input_path)
    output_folder_path = Path(configs.output_folder_path)
    image_folder_path = Path(configs.image_folder_path)

    s3 = S3(
        profile_name=configs.profile_name,
        aws_access_key=configs.aws_access_key,
        aws_secret_key=configs.aws_secret_key,
        bucket_name=configs.bucket_name,
        usekey=usekey,  # use profile or key
    )

    runner = create_obs3dian_runner(
        s3, image_folder_path, output_folder_path, overwrite
    )  # create main function

    if user_input_path.is_dir():
        markdown_file_paths = [
            markdown_path for markdown_path in user_input_path.rglob("**/*.md")
        ]  # get all .md file under input dir
    else:
        markdown_file_paths = [user_input_path]

    assert len(
        markdown_file_paths
    ), f"No md files in {user_input_path}"  # raise error when no md file in input path

    with concurrent_futures.ThreadPoolExecutor(
        max_workers=1
    ) as executor:  # thread for aniamtion
        with typer.progressbar(
            label="Processing", iterable=markdown_file_paths, show_eta=False
        ) as progeress:  # typer progress bar
            for markdown_file_path in progeress:
                # print progress bar
                typer.echo("")  # new line
                future = executor.submit(runner, markdown_file_path)

                _render_animation(future, f"{markdown_file_path.name}")
                future.result()  # check future result

                typer.echo(f"\rFinished    [{markdown_file_path.name}]")
                typer.echo("")  # new line

    typer.echo("\n")  # new line after progress bar
    typer.echo(
        f"Total converts: {len(markdown_file_paths)}\nobs3dian is successfully finished\n"
    )


if __name__ == "__main__":
    app()
